
// adapted from Folkert van Heusden <mail@vanheusden.com>
// Released under AGPL v3.0
// with code from https://github.com/PaulStoffregen/Time
// and http://jtauber.github.io/mars-clock/
// modified by Paul Dassonville <prd@me.com> under AGPL v.3.0

/*
After first uploading code on the ESP8266, clock motion and colors are a little glitchy.
To correct, complete a parameter setting, and then reboot the clock by unplugging/replugging.

To set up the Mars 2020 Clock:
* Open back cover
* Insert power cord into the plug, and plug into power outlet
* On initial setup, clock will automatically enter the setup routine as indicated by the clock rings turning YELLOW (after initial setup, the setup routine can be revisited by pressing the small button near the plug in the back of the clock)
* After clocks turn YELLOW, join the MARS_CLOCK WiFi network (password = Mars2020) using a mobile device or computer
* Navigate to http://192.168.4.1 in a browser, and click “Configure WiFi”
* Select a WiFi network and enter the password
* Adjust the other parameters for the clocks as desired
* Press the “Save” button to save updated parameters
* Navigate back to http://192.168.4.1, and click “Exit” to restart the clock
* Place back cover back on the clock, making sure that the cord runs through one of the slots near the lower corner of the frame
* Enjoy!

Full rings of color on the clock faces indicate:
YELLOW — Awaiting parameter setup
PURPLE — Awaiting WiFi connection
GREEN — Awaiting time server connection
BLUE — Awaiting time sync

For more information or assistance, email prd@me.com 
*/

#include <FS.h>  
#include <ArduinoJson.h>          //https://github.com/bblanchon/ArduinoJson
#include <TimeLib.h>
#include <ESP8266WiFi.h>
#include <WiFiUdp.h>
#include <Timezone.h>
#include <DNSServer.h>            //Local DNS Server used for redirecting all requests to the configuration portal
#include <ESP8266WebServer.h>     //Local WebServer used to serve the configuration portal
#include <WiFiManager.h>          //https://github.com/tzapu/WiFiManager WiFi Configuration Magic

bool debug = false;
//int buttonPin = 0;
char STDtimezone[7] = "PST";
char DSTtimezone[7] = "PDT";
float longitude; //longitude used to calculate Local Mean Solar Time on Mars
char longitudeVal[7] = "77.45";
float earthHours;
char earthHoursVal[3] = "24";
float marsHours;
char marsHoursVal[3] = "24";
char marsDisplayVal[6] = "BOTH";
//static const char ntpServerName[] = "73.158.5.1"; // "193.67.79.202"
char ntpServerNamePrev[16] = "73.158.5.1"; // "193.67.79.202"
char ntpServerName[16] = "73.158.5.1"; // "193.67.79.202"
// Difference between TAI and UTC (updated to 37 on 12/31/2016). This value should be updated each time the IERS announces a leap second.
double tai_offset = 37;
char tai_offsetVal[3] = "37";
float brightnessMax = 255.0;
float brightnessMaxProp = 100.0;
char brightnessMaxVal[4] = "100";
bool firstTime = true;

TimeChangeRule edt = {"EDT", Second, Sun, Mar, 2, -240};    // Daylight time = UTC - 4 hours, hence -240 min
TimeChangeRule est = {"EST", First, Sun, Nov, 2, -300};     // Standard time = UTC - 5 hours, hence -300 min
TimeChangeRule cdt = {"CDT", Second, Sun, Mar, 2, -300};    // Daylight time = UTC - 5 hours, hence -300 min
TimeChangeRule cst = {"CST", First, Sun, Nov, 2, -360};     // Standard time = UTC - 6 hours, hence -360 min
TimeChangeRule mdt = {"MDT", Second, Sun, Mar, 2, -360};    // Daylight time = UTC - 6 hours, hence -360 min
TimeChangeRule mst = {"MST", First, Sun, Nov, 2, -420};     // Standard time = UTC - 7 hours, hence -420 min
TimeChangeRule pdt = {"PDT", Second, Sun, Mar, 2, -420};    // Daylight time = UTC - 7 hours, hence -420 min
TimeChangeRule pst = {"PST", First, Sun, Nov, 2, -480};     // Standard time = UTC - 8 hours, hence -480 min
  
TimeChangeRule myDST = pdt;    // Daylight time = UTC - 7 hours, hence -420 min;
TimeChangeRule mySTD = pst;     // Standard time = UTC - 8 hours, hence -480 min;
Timezone myTZ(myDST, mySTD);
TimeChangeRule *tcr;        // pointer to the time change rule, use to get TZ abbrev
void setTimezone(String DSTtimezone, String STDtimezone) {
  if (DSTtimezone == "") DSTtimezone = STDtimezone;
  STDtimezone.toUpperCase();
  DSTtimezone.toUpperCase();
  if (STDtimezone.startsWith("UTC")) {
    STDtimezone.replace("UTC", "");
    int timeOffset = STDtimezone.toInt();
    TimeChangeRule newTime = {"UTC", First, Sun, Nov, 2, timeOffset * 60};
    mySTD = newTime;
  }  
  if (DSTtimezone.startsWith("UTC")) {
    DSTtimezone.replace("UTC", "");
    int timeOffset = DSTtimezone.toInt();
    TimeChangeRule newTime = {"UTC", First, Sun, Nov, 2, timeOffset * 60};
    myDST = newTime;
  } 
  if (DSTtimezone == "EDT") myDST = edt;    // Daylight time = UTC - 4 hours, hence -240 min
  if (STDtimezone == "EST") mySTD = est;     // Standard time = UTC - 5 hours, hence -300 min
  if (DSTtimezone == "CDT") myDST = cdt;    // Daylight time = UTC - 5 hours, hence -300 min
  if (STDtimezone == "CST") mySTD = cst;     // Standard time = UTC - 6 hours, hence -360 min
  if (DSTtimezone == "MDT") myDST = mdt;    // Daylight time = UTC - 6 hours, hence -360 min
  if (STDtimezone == "MST") mySTD = mst;     // Standard time = UTC - 7 hours, hence -420 min
  if (DSTtimezone == "PDT") myDST = pdt;    // Daylight time = UTC - 7 hours, hence -420 min
  if (STDtimezone == "PST") mySTD = pst;     // Standard time = UTC - 8 hours, hence -480 min
  myTZ = Timezone(myDST, mySTD);
}

     
unsigned long oldMillis = millis();
unsigned long newMillis = millis();
double newUtc = now();
double oldUtc = now();
unsigned long startMillis = millis();
byte msOld = 0;
unsigned long esLEDSwitch = 0;
int esLEDOld = 0;
unsigned long emLEDSwitch = 0;
int emLEDOld = 0;
unsigned long ehLEDSwitch = 0;
int ehLEDOld = 0;

#include <FastLED.h>
#define DATA_PIN_MARS    4
#define CLOCK_PIN_MARS   2
#define DATA_PIN_EARTH    7
#define CLOCK_PIN_EARTH   5
uint32_t mars_pixels = 60; //2 Inch
uint32_t earth_pixels = 40; //2 Inch
#define NUMBER_OF_PIXELS 100
#define NUM_LEDS 100
CRGB ring[NUM_LEDS];
CRGB color;
#define FRAMES_PER_SECOND  120

uint32_t pixelsRed[NUMBER_OF_PIXELS];
uint32_t pixelsGreen[NUMBER_OF_PIXELS];
uint32_t pixelsBlue[NUMBER_OF_PIXELS];

uint32_t mh_pixelsRed[NUMBER_OF_PIXELS];
uint32_t mh_pixelsGreen[NUMBER_OF_PIXELS];
uint32_t mh_pixelsBlue[NUMBER_OF_PIXELS];

uint32_t mm_pixelsRed[NUMBER_OF_PIXELS];
uint32_t mm_pixelsGreen[NUMBER_OF_PIXELS];
uint32_t mm_pixelsBlue[NUMBER_OF_PIXELS];

uint32_t ms_pixelsRed[NUMBER_OF_PIXELS];
uint32_t ms_pixelsGreen[NUMBER_OF_PIXELS];
uint32_t ms_pixelsBlue[NUMBER_OF_PIXELS];

uint32_t md_pixelsRed[NUMBER_OF_PIXELS];
uint32_t md_pixelsGreen[NUMBER_OF_PIXELS];
uint32_t md_pixelsBlue[NUMBER_OF_PIXELS];

uint32_t eh_pixelsRed[NUMBER_OF_PIXELS];
uint32_t eh_pixelsGreen[NUMBER_OF_PIXELS];
uint32_t eh_pixelsBlue[NUMBER_OF_PIXELS];

uint32_t em_pixelsRed[NUMBER_OF_PIXELS];
uint32_t em_pixelsGreen[NUMBER_OF_PIXELS];
uint32_t em_pixelsBlue[NUMBER_OF_PIXELS];

uint32_t es_pixelsRed[NUMBER_OF_PIXELS];
uint32_t es_pixelsGreen[NUMBER_OF_PIXELS];
uint32_t es_pixelsBlue[NUMBER_OF_PIXELS];

constexpr char ssid[] = "temp"; //PRD: Update for new wifi
constexpr char pass[] = "temp"; //PRD: Update for new wifi

WiFiUDP Udp;
unsigned int localPort = 8888;  // local port to listen for UDP packets

const int NTP_PACKET_SIZE = 48; // NTP time is in the first 48 bytes of message
byte packetBuffer[NTP_PACKET_SIZE]; //buffer to hold incoming & outgoing packets
bool NtpSuccess = false;
bool NtpError = false;
bool WifiError = false;

void setAllColor(uint8_t r, uint8_t g, uint8_t b) {
    for (int i = 0; i < NUM_LEDS; i++) ring[i] = CRGB(r,g,b);
    FastLED.show();
}

void getParams() {
    setAllColor(brightnessMax / 12, brightnessMax / 12, 0);
    WiFiManager wifiManager;
    WiFiManagerParameter ErrorInfo("<h2><font color=red>Error:</h2><font color=black>");
    WiFiManagerParameter WifiErrorInfo("<font color=red><h3>WiFi server not found. Please select a different WiFi server above.</h3><font color=black>");
    WiFiManagerParameter NTPErrorInfo("<font color=red><h3>Time server not found. Please select a different NTP server below. (This error could also be caused by a WiFi server that cannot be found.)</h3><font color=black>");
    WiFiManagerParameter earthInfo("<h2>Earth Clock</h2>");
    WiFiManagerParameter customSTDTimezone("stdOffset", "<strong>U.S. Standard Timezone or UTC with offset (EST, CST, MST, PST, UTC-1, UTC+0, UTC+1, etc.)</strong>", STDtimezone, 6);
    WiFiManagerParameter customDSTTimezone("dstOffset", "<p><strong>U.S. Daylight Saving Timezone (EDT, CDT, MDT, PDT)</strong><br>If left blank, will use Standard Timezone or UTC with offset for entire year", DSTtimezone, 6);
    WiFiManagerParameter customEarthHours("EarthHours", "<p><strong>12/24 Hr Clock (12 = 12 hr, 24 = 24 hr military time)</strong>", earthHoursVal, 2);
    WiFiManagerParameter marsInfo("<h2>Mars Clock</h2>");
    WiFiManagerParameter customMarsDisplay("marsDisplayVal", "<p><strong>Mars Display (CLOCK, TRIP, or BOTH)</strong><br>If CLOCK, displays Mars time at specified longitude. If TRIP, displays an arc depicting the proportion of Perseverance's trip duration completed (Earth to Mars). If BOTH, displays both Mars time and the trip completion. After Perseverance's estimated arrival time (Feb. 18, 2021, 19:43:38 UTC), display defaults to Mars time only", marsDisplayVal, 5);
    WiFiManagerParameter customMarsLongitude("MarsLongitude", "<strong>Mars Longitude (–180 to +180)</strong><br>Positive = East of meridian, negative = West of meridian<br>Use 0.0 for Coordinated Mars Time, approx. 77.45 for Perseverance, 137.4 for Curiosity, 135.6 for InSight, 175.5 for Spirit, –5.5 for Opportunity, –33.3 for Pathfinder, –125.8 for Phoenix, –48.0 for Viking 1, 134.3 for Viking 2", longitudeVal, 6);
    WiFiManagerParameter customMarsHours("MarsHours", "<p><strong>12/24 Hr Clock (12 = 12 hr, 24 = 24 hr miltary time)</strong>", marsHoursVal, 2);
    WiFiManagerParameter otherInfo("<h2>Other Parameters</h2>");
    WiFiManagerParameter customNTP("timeServer", "<strong>NTP Time Server IP (73.158.5.1 = Santa Clara, CA)</strong><br>If default time server not found, others are listed at https://support.ntp.org/bin/view/Servers/StratumOneTimeServers", ntpServerName, 16);
    WiFiManagerParameter customLeapSec("leapSec", "<strong>Leap Second Correction</strong><br>Difference between TAI (International Atomic Time) and UTC. Increment each time the IERS announces a leap second (updated to 37 on 12/31/2016)", tai_offsetVal, 2);
    WiFiManagerParameter customBrightness("brightnessMax", "<p><strong>Maximum Brightness (0 – 100)</strong>", brightnessMaxVal, 3);
// To change clock parameters, remove the back cover of the clock and press the small button on the ESP8266. When clock turns yellow then join the MARS_CLOCK wifi network (password = Mars2020); times out after 60 seconds if no connection.
// Go to http://192.168.4.1/wifi to set the wifi network that the clock will use to access the time server, and other clock parameters. Press SAVE when done.
// Go to http://192.168.4.1/info for more info.
// When changes are complete go to http://192.168.4.1/exit to exit and restart the clock.
    wifiManager.setConfigPortalTimeout(180);
    if (NtpError || WifiError) wifiManager.addParameter(&ErrorInfo);
    if (WifiError) wifiManager.addParameter(&WifiErrorInfo);
    if (NtpError) wifiManager.addParameter(&NTPErrorInfo);
    wifiManager.addParameter(&earthInfo);
    wifiManager.addParameter(&customSTDTimezone);
    wifiManager.addParameter(&customDSTTimezone);
    wifiManager.addParameter(&customEarthHours);
    wifiManager.addParameter(&marsInfo);
    wifiManager.addParameter(&customMarsDisplay);
    wifiManager.addParameter(&customMarsLongitude);
    wifiManager.addParameter(&customMarsHours);
    wifiManager.addParameter(&otherInfo);
    wifiManager.addParameter(&customNTP);
    wifiManager.addParameter(&customLeapSec);
    wifiManager.addParameter(&customBrightness);
    wifiManager.setSaveConfigCallback(saveConfigCallback);
  if (firstTime) {
    wifiManager.autoConnect("MARS_CLOCK", "Mars2020");
    //wifiManager.startConfigPortal("MARS_CLOCK", "Mars2020");
    Serial.println("connected...yeey :)");
  }

  if (!firstTime) {
    wifiManager.startConfigPortal("MARS_CLOCK", "Mars2020");
    Serial.println("connected...yeey :)");
  
    strcpy(STDtimezone, customSTDTimezone.getValue());
    strcpy(DSTtimezone, customDSTTimezone.getValue());
    setTimezone(DSTtimezone, STDtimezone);
    strcpy(earthHoursVal, customEarthHours.getValue());
    if (((earthHoursVal[0] == '1') && (earthHoursVal[1] == '2')) || ((earthHoursVal[0] == '2') && (earthHoursVal[1] == '4'))) {
      earthHours = atof(earthHoursVal);
    }
    else {
      strcpy(earthHoursVal, "24");
      earthHours = 24;
    }
    strcpy(longitudeVal, customMarsLongitude.getValue());
    longitude = atof(longitudeVal);
    strcpy(marsHoursVal, customMarsHours.getValue());
    if (((marsHoursVal[0] == '1') && (marsHoursVal[1] == '2')) || ((marsHoursVal[0] == '2') && (marsHoursVal[1] == '4'))) {
      marsHours = atof(marsHoursVal);
    }
    else {
      strcpy(marsHoursVal, "24");
      marsHours = 24;
    }
    strcpy(marsDisplayVal, customMarsDisplay.getValue());
    String str_marsDisplayVal(marsDisplayVal);
    str_marsDisplayVal.toUpperCase();
    strcpy(marsDisplayVal, str_marsDisplayVal.c_str());
    if (!str_marsDisplayVal.startsWith("TRIP") && !str_marsDisplayVal.startsWith("BOTH")) {
      strcpy(marsDisplayVal, "CLOCK");
    }
    strcpy(ntpServerName, customNTP.getValue());
    strcpy(tai_offsetVal, customLeapSec.getValue());
    tai_offset = atoi(tai_offsetVal);
    strcpy(brightnessMaxVal, customBrightness.getValue());
    brightnessMaxProp = atof(brightnessMaxVal);
    if (brightnessMaxProp > 100) {
      Serial.println("Adjusting brightnessMaxProp to 100");
      strcpy(brightnessMaxVal, "100");
      brightnessMaxProp = 100;
    }
    if (brightnessMaxProp < 0) {
      Serial.println("Adjusting brightnessMaxProp to 0");
      strcpy(brightnessMaxVal, "0");
      brightnessMaxProp = 0;
    }
    brightnessMax = 255 * brightnessMaxProp / 100;
    FastLED.setBrightness(brightnessMax);

    // Save parameters
    Serial.println("saving config");
    DynamicJsonDocument jsonBuffer(1024);
    jsonBuffer["STDtimezone"] = STDtimezone;
    jsonBuffer["DSTtimezone"] = DSTtimezone;
    jsonBuffer["earthHoursVal"] = earthHoursVal;
    jsonBuffer["longitudeVal"] = longitudeVal;
    jsonBuffer["marsHoursVal"] = marsHoursVal;
    jsonBuffer["marsDisplayVal"] = marsDisplayVal;
    jsonBuffer["ntpServerName"] = ntpServerName;
    jsonBuffer["tai_offsetVal"] = tai_offsetVal;
    jsonBuffer["brightnessMaxVal"] = brightnessMaxVal;
  
    File configFile = SPIFFS.open("/config.json", "w");
    if (!configFile) {
      Serial.println("failed to open config file for writing");
    }

    serializeJson(jsonBuffer, Serial);
    serializeJson(jsonBuffer, configFile);
    configFile.close();

    setAllColor(0, 0, 0);
    //delay(500);
  }
  firstTime = false;
}

// send an NTP request to the time server at the given address
void sendNTPpacket(IPAddress &address) {
  Serial.println("sendNTPpacket");
	// set all bytes in the buffer to 0
	memset(packetBuffer, 0, NTP_PACKET_SIZE);

	// Initialize values needed to form NTP request
	// (see URL above for details on the packets)
	packetBuffer[0] = 0b11100011;   // LI, Version, Mode
	packetBuffer[1] = 0;     // Stratum, or type of clock
	packetBuffer[2] = 6;     // Polling Interval
	packetBuffer[3] = 0xEC;  // Peer Clock Precision
	// 8 bytes of zero for Root Delay & Root Dispersion
	packetBuffer[12] = 49;
	packetBuffer[13] = 0x4E;
	packetBuffer[14] = 49;
	packetBuffer[15] = 52;

	// all NTP fields have been given values, now
	// you can send a packet requesting a timestamp:
	Udp.beginPacket(address, 123); //NTP requests are to port 123
	Udp.write(packetBuffer, NTP_PACKET_SIZE);
	Udp.endPacket();
}

time_t getNtpTime() {
	IPAddress ntpServerIP; // NTP server's ip address
	while (Udp.parsePacket() > 0) {
	} // discard any previously received packets
	Serial.println("Transmit NTP Request");
	// get a random server from the pool
	WiFi.hostByName(ntpServerName, ntpServerIP);
	Serial.print(ntpServerName);
	Serial.print(": ");
	Serial.println(ntpServerIP);
	sendNTPpacket(ntpServerIP);
	uint32_t beginWait = millis();
	while (millis() - beginWait < 30000) {
		int size = Udp.parsePacket();
		if (size >= NTP_PACKET_SIZE) {
			Serial.println("Receive NTP Response");
			Udp.read(packetBuffer, NTP_PACKET_SIZE);  // read packet into the buffer
			unsigned long secsSince1900;
			// convert four bytes starting at location 40 to a long integer
			secsSince1900 =  (unsigned long)packetBuffer[40] << 24;
			secsSince1900 |= (unsigned long)packetBuffer[41] << 16;
			secsSince1900 |= (unsigned long)packetBuffer[42] << 8;
			secsSince1900 |= (unsigned long)packetBuffer[43];
      NtpSuccess = true;
			return secsSince1900 - 2208988800UL;
		}
	}
	Serial.println("No NTP Response :-(");
  NtpSuccess = false;
	return 0; // return 0 if unable to get the time
}

void saveConfigCallback() {
  Serial.println('Resyncing NTP');
  setSyncProvider(getNtpTime);
}


void setup() {
  Serial.begin(115200);
  Serial.println(F("Init MARS_CLOCK " __DATE__ " " __TIME__));
  delay(3000);
  pinMode(D1, INPUT_PULLUP);

  Serial.println('Starting up');
  LEDS.addLeds<APA102, DATA_PIN_MARS, CLOCK_PIN_MARS, BGR>(ring, 0, mars_pixels);
  LEDS.addLeds<APA102, DATA_PIN_EARTH, CLOCK_PIN_EARTH, BGR>(ring, mars_pixels, earth_pixels);
  setAllColor(brightnessMax / 20, brightnessMax / 20, brightnessMax / 20);

  //clean FS, for testing
  //SPIFFS.format();

  Serial.println("mounting FS...");

  if (SPIFFS.begin()) {
    Serial.println("mounted file system");
    if (SPIFFS.exists("/config.json")) {
      //file exists, reading and loading
      Serial.println("reading config file");
      File configFile = SPIFFS.open("/config.json", "r");
      if (configFile) {
        Serial.println("opened config file");
        size_t size = configFile.size();
        // Allocate a buffer to store contents of the file.
        std::unique_ptr<char[]> buf(new char[size]);

        configFile.readBytes(buf.get(), size);
        DynamicJsonDocument jsonBuffer(1024);
        deserializeJson(jsonBuffer, buf.get());
        if (!jsonBuffer.isNull()) {
          Serial.println("\nparsed json");
          strcpy(STDtimezone, jsonBuffer["STDtimezone"]);
          strcpy(DSTtimezone, jsonBuffer["DSTtimezone"]);
          strcpy(earthHoursVal, jsonBuffer["earthHoursVal"]);
          earthHours = atof(earthHoursVal);
          strcpy(longitudeVal, jsonBuffer["longitudeVal"]);
          longitude = atof(longitudeVal);
          strcpy(marsHoursVal, jsonBuffer["marsHoursVal"]);
          marsHours = atof(marsHoursVal);
          strcpy(marsDisplayVal, jsonBuffer["marsDisplayVal"]);
          strcpy(ntpServerName, jsonBuffer["ntpServerName"]);
          strcpy(ntpServerNamePrev, ntpServerName);
          strcpy(tai_offsetVal, jsonBuffer["tai_offsetVal"]);
          tai_offset = atoi(tai_offsetVal);
          strcpy(brightnessMaxVal, jsonBuffer["brightnessMaxVal"]);
          brightnessMaxProp = atof(brightnessMaxVal);
          brightnessMax = 255 * brightnessMaxProp /100;
        }
        else {
          Serial.println("failed to load json config");
        }
        configFile.close();
      }
    }
  } else {
    Serial.println("failed to mount FS");
  }

  setAllColor(brightnessMax / 12, brightnessMax / 12, 0);
  getParams();

	Serial.print("IP address assigned by DHCP is ");
	Serial.println(WiFi.localIP());

	setAllColor(brightnessMax / 12, 0, brightnessMax / 12);
	Serial.println("Starting UDP");
	Udp.begin(localPort);
	Serial.print("Local port: ");
	Serial.println(Udp.localPort());
  bool NtpReady = false;
	while (!NtpReady) {
    Serial.println();
    Serial.println("waiting for sync");
    setAllColor(0, brightnessMax / 8, 0);
    delay(500);
 	  setSyncProvider(getNtpTime);
    if (!NtpSuccess) {
      NtpError = true;
      getParams();
      uint32_t beginWait = millis();
      Serial.print("Reconnecting to wifi");
      setAllColor(brightnessMax / 12, 0, brightnessMax / 12);
      while ((millis() - beginWait < 30000) && (WiFi.status() != WL_CONNECTED)) {
        delay(500);
        Serial.print(".");
      }
      Serial.println();
      if (WiFi.status() != WL_CONNECTED) WifiError = true;
      else WifiError = false;
    }
    else {
      NtpError = false;
      NtpReady = true;
    }
  }

	setSyncInterval(86400); // 86400 = time synced every 24 hr (86400 sec)

	setAllColor(0, 0, brightnessMax / 8);
  newUtc = now();
  oldUtc = now();
  while (newUtc <= oldUtc) newUtc = now(); 
  startMillis = millis();

	Serial.println(F("Go!"));
  setAllColor(0, 0, 0);
}

void loop() {
  bool buttonVal = digitalRead(D1);
  if (!buttonVal) {
    Serial.println("button pressed");
    getParams();
    Serial.print("ntpServerNamePrev: ");
    Serial.println(ntpServerNamePrev);
    Serial.print("ntpServerName: ");
    Serial.println(ntpServerName);
    if (strcmp(ntpServerNamePrev, ntpServerName) != 0) {
      bool NtpReady = false; 

      //uint32_t beginWait = millis();
      //Serial.print("Reconnecting to wifi");
      //setAllColor(brightnessMax / 12, 0, brightnessMax / 12);
      //while ((millis() - beginWait < 30000) && (WiFi.status() != WL_CONNECTED)) {
      //  delay(500);
      //  Serial.print(".");
      //}
      //if (WiFi.status() != WL_CONNECTED) WifiError = true;
      //else WifiError = false;
      
      while (!NtpReady) {
        uint32_t beginWait = millis();
        Serial.print("Reconnecting to wifi");
        setAllColor(brightnessMax / 12, 0, brightnessMax / 12);
        while ((millis() - beginWait < 30000) && (WiFi.status() != WL_CONNECTED)) {
          delay(500);
          Serial.print(".");
        }
        Serial.println();
        if (WiFi.status() != WL_CONNECTED) WifiError = true;
        else WifiError = false;
        Serial.println("waiting for sync");
        setAllColor(0, brightnessMax / 8, 0);
        delay(500);
        setSyncProvider(getNtpTime);

        if (!NtpSuccess) {
          NtpError = true;
          getParams();
        }
        else{
          NtpError = false;
          NtpReady = true;
        }
      }
 
      setSyncInterval(86400); // 86400 = time synced every 24 hr (86400 sec)
      strcpy(ntpServerNamePrev, ntpServerName);

      setAllColor(0, 0, brightnessMax / 8);
      newUtc = now();
      oldUtc = now();
      while (newUtc <= oldUtc) newUtc = now(); 
      startMillis = millis();
    }
    Serial.println(F("Go!"));
    setAllColor(0, 0, 0);
  }

  newMillis = millis();
  while (newMillis < oldMillis+100) newMillis = millis();
  oldMillis = newMillis;
  //time_t utc = now();
  double utc = now();
    if (oldUtc < utc) {
    oldUtc = utc;
    startMillis = newMillis;
    //Serial.println("New Second");
    }
  time_t local = myTZ.toLocal(utc, &tcr);
  utc = utc + (newMillis-startMillis)/1000.0;
  double es = second(local) + (utc - floor(utc));
  double em = minute(local) + es/60.0;
  double eh = hour(local) + em/60.0;
//  es = 0.0; // for testing only -- comment out
//  em = 0.0; // for testing only -- comment out
//  eh = 0.0; // for testing only -- comment out

  double launchTime = 1596109800; // July 30, 2020, 11:50:00, 1596109800
  double arrivalTime = 1613677418; // Feb 18, 2021, 19:43:38, 1613677418
  double tripProp = (utc - launchTime)/(arrivalTime - launchTime);
  Serial.println(now());
  Serial.println(tripProp);
  
	double jd_ut = 2440587.5 + (utc / 8.64E4);
	double jd_tt = jd_ut + (tai_offset + 32.184) / 86400;
	double j2000 = jd_tt - 2451545.0;
	double m = fmod(19.3870 + 0.52402075 * j2000, 360.0);
	double alpha_fms = fmod(270.3863 + 0.52403840 * j2000, 360.0);
	double e = 0.09340 + 2.477E-9 * j2000;
	double pbs =
		0.0071 * cos((0.985626 * j2000 /  2.2353) +  49.409) +
		0.0057 * cos((0.985626 * j2000 /  2.7543) + 168.173) +
		0.0039 * cos((0.985626 * j2000 /  1.1177) + 191.837) +
		0.0037 * cos((0.985626 * j2000 / 15.7866) +  21.736) +
		0.0021 * cos((0.985626 * j2000 /  2.1354) +  15.704) +
		0.0020 * cos((0.985626 * j2000 /  2.4694) +  95.528) +
		0.0018 * cos((0.985626 * j2000 / 32.8493) +  49.095);
	double nu_m = (10.691 + 3.0E-7 * j2000) * sin(m) +
		0.623 * sin(2 * m) +
		0.050 * sin(3 * m) +
		0.005 * sin(4 * m) +
		0.0005 * sin(5 * m) +
		pbs;
	double nu = nu_m + m;
	double l_s = fmod(alpha_fms + nu_m, 360.0);
	double eot = 2.861 * sin(2 * l_s) - 0.071 * sin(4 * l_s) + 0.002 * sin(6 * l_s) - nu_m;
	double eot_h = eot * 24 / 360;
	double msd = (((j2000 - 4.5) / 1.027491252) + 44796.0 - 0.00096);
	double mtc = fmod(24.0 * msd, 24.0);  // Coordinated Mars Time (mean solar time at Mars's prime meridian)
  double lmst = fmod(24.0 * msd + longitude / 15.0, 24.0); // Local Mean Solar Time


	double x = lmst * 3600.0;
	double y = fmod(x, 3600.0);
  double mm = y / 60.0;
  double ms = (fmod(y, 60.0));
  double mh = floor(lmst) + mm/60.0;
//  mm = 0.0; // for testing only -- comment out
//  ms = 0.0; // for testing only -- comment out
//  mh = 0.0; // for testing only -- comment out



  if (debug) {
    Serial.print("Mars ");
    Serial.print(mh);
    Serial.print(':');
    Serial.print(mm);
    Serial.print(':');
    Serial.println(ms);
  }
  
  if (mh >= 12.0 & marsHours == 12) mh = mh - 12;
  double mhLEDFraction = mh * mars_pixels / marsHours;
  int mhLED = floor(mhLEDFraction); 
  int mhLEDNext = mhLED + 1;
  if (mhLEDNext == mars_pixels) mhLEDNext = 0;
  double mhBrightFraction = mhLEDFraction - floor(mhLEDFraction);
  double mhBright = 255.0 - 255.0 * mhBrightFraction;
  double mhNextBright = 255.0 * mhBrightFraction;
  if (debug) {
    Serial.println("Mars hour breakdown");
    Serial.print(mhLED);
    Serial.print(' ');
    Serial.println(mhBright);
    Serial.print(mhLEDNext);
    Serial.print(' ');
    Serial.println(mhNextBright);
    Serial.println(mhBrightFraction);
  }
  
  double mmLEDFraction = mm * mars_pixels / 60.0;
  int mmLED = floor(mmLEDFraction);
  int mmLEDNext = mmLED + 1;
  if (mmLEDNext == mars_pixels) mmLEDNext = 0;
  double mmBrightFraction = mmLEDFraction - floor(mmLEDFraction);
  double mmBright = 255.0 - 255.0 * mmBrightFraction;
  double mmNextBright = 255.0 * mmBrightFraction;
  if (debug) {
    Serial.println("Mars minute breakdown");
    Serial.print(mmLED);
    Serial.print(' ');
    Serial.println(mmBright);
    Serial.print(mmLEDNext);
    Serial.print(' ');
    Serial.println(mmNextBright);
    Serial.println(mmBrightFraction);
  }
  
  double msLEDFraction = ms * mars_pixels / 60.0;
  int msLED = floor(msLEDFraction);
  int msLEDNext = msLED + 1;
  if (msLEDNext == mars_pixels) msLEDNext = 0;
  double msBrightFraction = msLEDFraction - floor(msLEDFraction);
  double msBright = 255.0 - 255.0 * msBrightFraction;
  double msNextBright = 255.0 * msBrightFraction;
  if (debug) {
    Serial.println("Mars second breakdown");
    Serial.print(msLED);
    Serial.print(' ');
    Serial.print(msBright);
    Serial.print(' ');
    Serial.println(msBrightFraction);
    Serial.print(msLEDNext);
    Serial.print(' ');
    Serial.println(msNextBright);
    Serial.println(msBrightFraction);
  }

  if (debug) {
    Serial.print("Earth ");
    Serial.print(eh);
    Serial.print(':');
    Serial.print(em);
    Serial.print(':');
    Serial.println(es);
  }
   
  if (eh >= 12.0 & earthHours == 12) eh = eh - 12;
  double ehLEDFraction = eh * earth_pixels / earthHours;
  int ehLED = floor(ehLEDFraction);
  int ehLEDNext = ehLED + 1;
  if (ehLEDNext == earth_pixels) ehLEDNext = 0;
  double ehBrightFraction = ehLEDFraction - floor(ehLEDFraction);
  double ehBright = 255.0 - 255.0 * ehBrightFraction;
  double ehNextBright = 255.0 * ehBrightFraction;
  if (debug) {
    Serial.println("Earth hour breakdown");
    Serial.print(ehLED);
    Serial.print(' ');
    Serial.println(ehBright);
    Serial.print(ehLEDNext);
    Serial.print(' ');
    Serial.println(ehNextBright);
    Serial.println(ehBrightFraction);
  }
  
  double emLEDFraction = em * earth_pixels / 60.0;
  int emLED = floor(emLEDFraction);
  int emLEDNext = emLED + 1;
  if (emLEDNext == earth_pixels) emLEDNext = 0;
  double emBrightFraction = emLEDFraction - floor(emLEDFraction);
  double emBright = 255.0 - 255.0 * emBrightFraction;
  double emNextBright = 255.0 * emBrightFraction;
  if (debug) {
    Serial.println("Earth minute breakdown");
    Serial.print(emLED);
    Serial.print(' ');
    Serial.println(emBright);
    Serial.print(emLEDNext);
    Serial.print(' ');
    Serial.println(emNextBright);
    Serial.println(emBrightFraction);
  }
  
  double esLEDFraction = es * earth_pixels / 60.0;
  int esLED = floor(esLEDFraction);
  int esLEDNext = esLED + 1;
  if (esLEDNext == earth_pixels) esLEDNext = 0;
  double esBrightFraction = esLEDFraction - floor(esLEDFraction);
  double esBright = 255.0 - 255.0 * esBrightFraction;
  double esNextBright = 255.0 * esBrightFraction;
  if (debug) {
    Serial.println("Earth second breakdown");
    Serial.print(esLED);
    Serial.print(' ');
    Serial.println(esBright);
    Serial.print(esLEDNext);
    Serial.print(' ');
    Serial.println(esNextBright);
    Serial.println(esBrightFraction);
  }

  
  for (uint8_t i = 0; i < NUMBER_OF_PIXELS; i++)   {
    pixelsRed[i] = 0;
    pixelsGreen[i] = 0;
    pixelsBlue[i] = 0;
    mh_pixelsRed[i] = 0;
    mm_pixelsRed[i] = 0;
    mm_pixelsGreen[i] = 0;
    ms_pixelsRed[i] = 0;
    ms_pixelsGreen[i] = 0;
    ms_pixelsBlue[i] = 0;
    eh_pixelsBlue[i] = 0;
    em_pixelsBlue[i] = 0;
    em_pixelsGreen[i] = 0;
    es_pixelsBlue[i] = 0;
    es_pixelsGreen[i] = 0;
    es_pixelsRed[i] = 0;
  }

    if ((strcmp(marsDisplayVal, "CLOCK") == 0) || (strcmp(marsDisplayVal, "BOTH") == 0)) {
      mh_pixelsRed[mhLED] = ceil(108 * mhBright/255); //108
      mh_pixelsRed[mhLEDNext] = ceil(108 * mhNextBright/255); //108
      mm_pixelsRed[mmLED] = ceil(55 * mmBright/255); //83
      mm_pixelsRed[mmLEDNext] = ceil(55 *mmNextBright/255); //83
      mm_pixelsGreen[mmLED] = ceil(25 * mmBright/255); //25
      mm_pixelsGreen[mmLEDNext] = ceil(25 * mmNextBright/255); //25
      ms_pixelsRed[msLED] = ceil(64 * msBright/255); //64
      ms_pixelsRed[msLEDNext] = ceil(64 * msNextBright/255); //64
      ms_pixelsGreen[msLED] = ceil(37 * msBright/255); //37
      ms_pixelsGreen[msLEDNext] = ceil(37 * msNextBright/255); //37
      ms_pixelsBlue[msLED] = ceil(7 * msBright/255); //7
      ms_pixelsBlue[msLEDNext] = ceil(7 * msNextBright/255); //7
    }   
    else {
      mh_pixelsRed[mhLED] = 0;
      mh_pixelsRed[mhLEDNext] = 0;
      mm_pixelsRed[mmLED] = 0;
      mm_pixelsRed[mmLEDNext] = 0;
      mm_pixelsGreen[mmLED] = 0;
      mm_pixelsGreen[mmLEDNext] = 0;
      ms_pixelsRed[msLED] = 0;
      ms_pixelsRed[msLEDNext] = 0;
      ms_pixelsGreen[msLED] = 0;
      ms_pixelsGreen[msLEDNext] = 0;
      ms_pixelsBlue[msLED] = 0;
      ms_pixelsBlue[msLEDNext] = 0;     
    }

    if ((utc < arrivalTime) && ((strcmp(marsDisplayVal, "TRIP") == 0) || (strcmp(marsDisplayVal, "BOTH") == 0))) {
      for (uint8_t i = 0; i < floor(tripProp*60); i++)   {
        md_pixelsRed[i] = 2;
        md_pixelsGreen[i] = 1;
        md_pixelsBlue[i] = 0;
      }
    }
    else {
      for (uint8_t i = 0; i < 60; i++)   {
        md_pixelsRed[i] = 0;
        md_pixelsGreen[i] = 0;
        md_pixelsBlue[i] = 0;
      }      
    }


    eh_pixelsBlue[ehLED+mars_pixels] = ceil(140 * ehBright/255);
    eh_pixelsBlue[ehLEDNext+mars_pixels] = ceil(140 * ehNextBright/255);
    em_pixelsBlue[emLED+mars_pixels] = ceil(60 * emBright/255);
    em_pixelsBlue[emLEDNext+mars_pixels] = ceil(60 *emNextBright/255);
    em_pixelsGreen[emLED+mars_pixels] = ceil(80 * emBright/255);
    em_pixelsGreen[emLEDNext+mars_pixels] = ceil(80 * emNextBright/255);
    es_pixelsBlue[esLED+mars_pixels] = ceil(55 * esBright/255);
    es_pixelsBlue[esLEDNext+mars_pixels] = ceil(55 * esNextBright/255);
    es_pixelsGreen[esLED+mars_pixels] = ceil(55 * esBright/255);
    es_pixelsGreen[esLEDNext+mars_pixels] = ceil(55 * esNextBright/255);
    es_pixelsRed[esLED+mars_pixels] = ceil(30 * esBright/255);
    es_pixelsRed[esLEDNext+mars_pixels] = ceil(30 * esNextBright/255);

    for (uint8_t i = 0; i < NUMBER_OF_PIXELS; i++)   {
     pixelsRed[i] = (mh_pixelsRed[i] + mm_pixelsRed[i] + ms_pixelsRed[i] + eh_pixelsRed[i] + em_pixelsRed[i] + es_pixelsRed[i] + md_pixelsRed[i]);
     pixelsGreen[i] = (mh_pixelsGreen[i] + mm_pixelsGreen[i] + ms_pixelsGreen[i] + eh_pixelsGreen[i] + em_pixelsGreen[i] + es_pixelsGreen[i] + md_pixelsGreen[i]);
     pixelsBlue[i] = (mh_pixelsBlue[i] + mm_pixelsBlue[i] + ms_pixelsBlue[i] + eh_pixelsBlue[i] + em_pixelsBlue[i] + es_pixelsBlue[i] + md_pixelsBlue[i]);
    }
  
 /*    ring[int((mars_pixels-mhLED)-floor((mars_pixels-mhLED)/mars_pixels)*mars_pixels)] = CRGB(pixelsRed[mhLED], pixelsGreen[mhLED], pixelsBlue[mhLED]);
    ring[int((mars_pixels-mhLEDNext)-floor((mars_pixels-mhLEDNext)/mars_pixels)*mars_pixels)] = CRGB(pixelsRed[mhLEDNext], pixelsGreen[mhLEDNext], pixelsBlue[mhLEDNext]);
    ring[int((mars_pixels-mmLED)-floor((mars_pixels-mmLED)/mars_pixels)*mars_pixels)] = CRGB(pixelsRed[mmLED], pixelsGreen[mmLED], pixelsBlue[mmLED]);
    ring[int((mars_pixels-mmLEDNext)-floor((mars_pixels-mmLEDNext)/mars_pixels)*mars_pixels)] = CRGB(pixelsRed[mmLEDNext], pixelsGreen[mmLEDNext], pixelsBlue[mmLEDNext]);
    ring[int((mars_pixels-msLED)-floor((mars_pixels-msLED)/mars_pixels)*mars_pixels)] = CRGB(pixelsRed[msLED], pixelsGreen[msLED], pixelsBlue[msLED]);
    ring[int((mars_pixels-msLEDNext)-floor((mars_pixels-msLEDNext)/mars_pixels)*mars_pixels)] = CRGB(pixelsRed[msLEDNext], pixelsGreen[msLEDNext], pixelsBlue[msLEDNext]);
*/
    for (uint8_t i = 0; i < mars_pixels; i++)   {
      ring[int((mars_pixels-i)-floor((mars_pixels-i)/mars_pixels)*mars_pixels)] = CRGB(pixelsRed[i], pixelsGreen[i], pixelsBlue[i]);
    }

    ring[int((earth_pixels-ehLED)-floor((earth_pixels-ehLED)/earth_pixels)*earth_pixels)+mars_pixels] = CRGB(pixelsRed[ehLED+mars_pixels], pixelsGreen[ehLED+mars_pixels], pixelsBlue[ehLED+mars_pixels]);
    ring[int((earth_pixels-ehLEDNext)-floor((earth_pixels-ehLEDNext)/earth_pixels)*earth_pixels)+mars_pixels] = CRGB(pixelsRed[ehLEDNext+mars_pixels], pixelsGreen[ehLEDNext+mars_pixels], pixelsBlue[ehLEDNext+mars_pixels]);
    ring[int((earth_pixels-emLED)-floor((earth_pixels-emLED)/earth_pixels)*earth_pixels)+mars_pixels] = CRGB(pixelsRed[emLED+mars_pixels], pixelsGreen[emLED+mars_pixels], pixelsBlue[emLED+mars_pixels]);
    ring[int((earth_pixels-emLEDNext)-floor((earth_pixels-emLEDNext)/earth_pixels)*earth_pixels)+mars_pixels] = CRGB(pixelsRed[emLEDNext+mars_pixels], pixelsGreen[emLEDNext+mars_pixels], pixelsBlue[emLEDNext+mars_pixels]);
    ring[int((earth_pixels-esLED)-floor((earth_pixels-esLED)/earth_pixels)*earth_pixels)+mars_pixels] = CRGB(pixelsRed[esLED+mars_pixels], pixelsGreen[esLED+mars_pixels], pixelsBlue[esLED+mars_pixels]);
    ring[int((earth_pixels-esLEDNext)-floor((earth_pixels-esLEDNext)/earth_pixels)*earth_pixels)+mars_pixels] = CRGB(pixelsRed[esLEDNext+mars_pixels], pixelsGreen[esLEDNext+mars_pixels], pixelsBlue[esLEDNext+mars_pixels]);

    FastLED.show();
    //FastLED.delay(100);

    ring[int((mars_pixels-mhLED)-floor((mars_pixels-mhLED)/mars_pixels)*mars_pixels)] = CRGB(0, 0, 0);
    ring[int((mars_pixels-mhLEDNext)-floor((mars_pixels-mhLEDNext)/mars_pixels)*mars_pixels)] = CRGB(0, 0, 0);
    ring[int((mars_pixels-mmLED)-floor((mars_pixels-mmLED)/mars_pixels)*mars_pixels)] = CRGB(0, 0, 0);
    ring[int((mars_pixels-mmLEDNext)-floor((mars_pixels-mmLEDNext)/mars_pixels)*mars_pixels)] = CRGB(0, 0, 0);
    ring[int((mars_pixels-msLED)-floor((mars_pixels-msLED)/mars_pixels)*mars_pixels)] = CRGB(0, 0, 0);
    ring[int((mars_pixels-msLEDNext)-floor((mars_pixels-msLEDNext)/mars_pixels)*mars_pixels)] = CRGB(0, 0, 0);

    ring[int((earth_pixels-ehLED)-floor((earth_pixels-ehLED)/earth_pixels)*earth_pixels)+mars_pixels] = CRGB(0, 0, 0);
    ring[int((earth_pixels-ehLEDNext)-floor((earth_pixels-ehLEDNext)/earth_pixels)*earth_pixels)+mars_pixels] = CRGB(0, 0, 0);
    ring[int((earth_pixels-emLED)-floor((earth_pixels-emLED)/earth_pixels)*earth_pixels)+mars_pixels] = CRGB(0, 0, 0);
    ring[int((earth_pixels-emLEDNext)-floor((earth_pixels-emLEDNext)/earth_pixels)*earth_pixels)+mars_pixels] = CRGB(0, 0, 0);
    ring[int((earth_pixels-esLED)-floor((earth_pixels-esLED)/earth_pixels)*earth_pixels)+mars_pixels] = CRGB(0, 0, 0);
    ring[int((earth_pixels-esLEDNext)-floor((earth_pixels-esLEDNext)/earth_pixels)*earth_pixels)+mars_pixels] = CRGB(0, 0, 0);


}
